''' This module implements the command line interface for tgcf,
using the modern and robust `typer`.
'''

import os
import asyncio
from enum import Enum
import logging
from typing import Optional

import typer
from dotenv import load_dotenv

from tgcf import __version__


load_dotenv('.env')

FAKE = bool(os.getenv('FAKE'))
app = typer.Typer(add_completion=False)


class Mode(str, Enum):
    '''tgcf works in two modes'''
    past = 'past'
    live = 'live'


def version_callback(value: bool):
    if value:
        print(__version__)
        raise typer.Exit()


def verbosity_callback(value: bool):
    if value:
        logging.info(
            'Verbosity turned on. \nThis is suitable for debugging.\n')
        level = logging.INFO
    else:
        level = logging.WARNING
    logging.basicConfig(level=level)


@app.command()
def main(
        mode: Mode = typer.Argument(...,
                                    help='Choose the mode in which you want to run tgcf.',
                                    envvar='TGCF_MODE'),
        verbose: Optional[bool] = typer.Option(None,
                                               '--loud', '-l',
                                               callback=verbosity_callback,
                                               envvar='LOUD',
                                               help='Increase output verbosity.'),

        version: Optional[bool] = typer.Option(None,
                                               '--version',
                                               '-v',
                                               callback=version_callback,
                                               help='Show version and exit.')

):
    ''' tgcf is a powerful tool for forwarding telegram messages from source to destination.

    Don't forget to star  https://github.com/aahnik/tgcf

    Telegram Channel https://telegram.me/tg_cf
    '''

    if FAKE:
        print(f'mode = {mode}')
        quit(1)

    if mode == mode.past:
        from tgcf.past import forward_job
        asyncio.run(forward_job())
    else:
        from tgcf.live import start_sync
        start_sync()


# AAHNIK 2021



# a custom config parser

import os
from typing import List, Optional

import yaml
from pydantic import BaseModel
from telethon.sessions import StringSession

CONFIG_FILE = 'tgcf.config.yml'


class Forward(BaseModel):
    source: int
    dest: List[int]
    offset: Optional[int] = 0


class Config(BaseModel):
    forwards: List[Forward]


def read_config():
    with open(CONFIG_FILE) as file:
        config_dict = yaml.full_load(file)
        try:
            config = Config(**config_dict)
        except Exception as err:
            print(err)
            quit(1)
        else:
            return config


def update_config(config: Config):
    with open(CONFIG_FILE, 'w') as file:
        yaml.dump(config.dict(), file)


def env_var(name: str, optional=False):
    var = os.getenv(name)

    while not var:
        if optional:
            break
        var = input(f'Enter {name}: ')
    return var


API_ID = env_var('API_ID')
API_HASH = env_var('API_HASH')
USERNAME = env_var('USERNAME', optional=True)
SESSION_STRING = env_var('SESSION_STRING', optional=True)

if SESSION_STRING:
    SESSION = StringSession(SESSION_STRING)
else:
    SESSION = 'tgcf'

CONFIG = read_config()



import asyncio
import logging

from telethon import TelegramClient
from telethon.tl.patched import MessageService
from telethon.errors.rpcerrorlist import FloodWaitError

from tgcf.config import CONFIG, API_ID, API_HASH, SESSION


async def forward_job():
    ''' The function that does the job of forwarding all existing messages in the concerned chats'''

    async with TelegramClient(SESSION, API_ID, API_HASH) as client:

        for forward in CONFIG.forwards:
            last_id = 0
            print(forward.source)
            print(forward.dest)
            async for message in client.iter_messages(forward.source,
                                                      reverse=True,
                                                      offset_id=forward.offset):
                if isinstance(message, MessageService):
                    continue
                try:
                    for destination in forward.dest:
                        await client.send_message(destination, message)
                    last_id = str(message.id)
                    logging.info('forwarding message with id = %s', last_id)
                    forward.offset = last_id
                except FloodWaitError as fwe:
                    print(f'Sleeping for {fwe}')
                    await asyncio.sleep(delay=fwe.seconds)
                except Exception as err:
                    logging.exception(err)
                    break



from telethon import events, TelegramClient
from tgcf.config import CONFIG, API_HASH, API_ID, SESSION


from_to = {}

KEEP_LAST_MANY = 10000

existing_hashes = []

_stored = {}


for forward in CONFIG.forwards:
    from_to[forward.source] = forward.dest


class EventUid:
    def __init__(self, event):
        self.chat_id = event.chat_id
        try:
            self.msg_id = event.id
        except:
            self.msg_id = event.deleted_id

    def __str__(self):
        return f'chat={self.chat_id} msg={self.msg_id}'

    def __eq__(self, other):
        return self.chat_id == other.chat_id and self.msg_id == other.msg_id

    def __hash__(self) -> int:
        return hash(self.__str__())


async def new_message_handler(event):
    chat_id = event.chat_id

    if chat_id not in from_to:
        return

    message = event.message

    global _stored

    event_uid = EventUid(event)

    length = len(_stored)
    exceeding = length - KEEP_LAST_MANY

    if exceeding > 0:
        for key in _stored:
            del _stored[key]
            break

    to_send_to = from_to.get(chat_id)

    if to_send_to:
        if event_uid not in _stored:
            _stored[event_uid] = []

        for recipient in to_send_to:
            fwded_msg = await event.client.send_message(recipient, message)
            _stored[event_uid].append(fwded_msg)

    existing_hashes.append(hash(message.text))


async def edited_message_handler(event):
    message = event.message

    chat_id = event.chat_id
    if chat_id in from_to:

        event_uid = EventUid(event)

        fwded_msgs = _stored.get(event_uid)
        if fwded_msgs:
            for msg in fwded_msgs:
                await msg.edit(message.text)
            return
        else:
            to_send_to = from_to.get(event.chat_id)
            for recipient in to_send_to:
                await event.client.send_message(recipient, message)


async def deleted_message_handler(event):
    chat_id = event.chat_id
    if chat_id in from_to:
        event_uid = EventUid(event)
        fwded_msgs = _stored.get(event_uid)
        if fwded_msgs:
            for msg in fwded_msgs:
                await msg.delete()
            return


ALL_EVENTS = {
    'new': (new_message_handler, events.NewMessage()),
    'edited': (edited_message_handler, events.MessageEdited()),
    'deleted': (deleted_message_handler, events.MessageDeleted())
}


def start_sync():
    client = TelegramClient(SESSION, API_ID, API_HASH)
    for key, val in ALL_EVENTS.items():
        print(f'Adding event {key}')
        client.add_event_handler(*val)
    client.start()
    client.run_until_disconnected()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tgcf', 'tgcf.extensions']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.1.2,<9.0.0',
 'PyYAML>=5.4.1,<6.0.0',
 'Telethon>=1.20,<2.0',
 'aiohttp>=3.7.4,<4.0.0',
 'cryptg>=0.2.post2,<0.3',
 'hachoir>=3.1.2,<4.0.0',
 'pydantic>=1.8.1,<2.0.0',
 'python-dotenv>=0.15.0,<0.16.0',
 'requests>=2.25.1,<3.0.0',
 'typer>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['tgcf = tgcf:cli.app']}

setup_kwargs = {
    'name': 'tgcf',
    'version': '0.1.6',
    'description': 'Coming soon...',
    'long_description': None,
    'author': 'aahnik',
    'author_email': 'daw@aahnik.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
